        ; programm for control L2 cache on Intel P6+ CPU's
        ; Intel processors with CPUID 61xh and more modern are supported
        ; Note: disabling L2 cache on Pentium PRO CPU is "pseudo",
        ; L2 will be detected, but really not work

org 100h
use16

MSR_BBL_CR_CTL =  0x119 ; L2 CACHE CONTROL REGISTER 
MSR_BBL_CR_CTL3 = 0x11E ; L2 CACHE CONTROL REGISTER 3 (Pentium II+ only)

CR0_NW =  0x20000000		; No Write Trough flag
CR0_CD =  0x40000000		; Cache disable flag

start:
        ; get command line argument
    mov al, byte[80h]
    or al, al; if length of string is equal zero, show usage and exit
    jz show_usage_msg ;if command-line arguments were not found, print usage and exit

    movzx cx, byte[80h]		;get length of string to cx
    mov si, 81h			; get addr of first letter to cx
    cld

cmdloop:
    lodsb
    cmp al, ' '			; if space, loop
    jz  cmdloop

    cmp al, 0dh			; if 0dh exit
    jz show_usage_msg 

    cmp al, 0			; if 0 exit
    jz show_usage_msg

    or al, 20h			; convrert symbol to lower case

    cmp al, 'c'
    je cpu_chk

    cmp al, 'e'
    je  cpu_chk

    cmp al, 'd'
    je cpu_chk

    cmp al, 'a'
    je cpu_chk


    loop cmdloop

                ; if letters c/e/d were not found, to print usage and exit
    mov ah, 09h ; show message "usage"
    mov dx, usage_str
    int 21h
    jmp exit

cpu_chk:                        ; checking CPU type
    mov byte[cmd_symbol], al    ; save cmd symbol

    mov eax, 00h        	; check vendor string
    cpuid
    cmp ebx, 756e6547h
    jne non_intel
    cmp ecx, 6c65746eh
    jne non_intel
    cmp edx, 49656e69h
    je  chk_cpu_model

non_intel:
    mov ah, 09h 		; show message "non-Intel CPU detected!"
    mov dx, non_intel_str
    int 21h
    jmp exit

chk_cpu_model:
    mov eax, 01h
    cpuid
    and ax, 0ff0h
    cmp ah, 0f0h
    je  p4			; found Pentium 4 CPU
    cmp ax, 0610h
    jb unsupported_cpu
    jne cmd_act			; Pentium II+ CPU was detected
    mov byte[ppro_cpu], 1     ; Pentium PRO CPU was detected
    jmp short cmd_act

unsupported_cpu:
    mov ah, 09h ; show message "CPU is unsupported!"
    mov dx, cpu_unsupport_str
    int 21h
    jmp exit

p4:    
    mov ah, 09h ; show message "Net Burst class (Pentium 4) CPU found!"
    mov dx, p4_cpu_str
    int 21h
    jmp exit

cmd_act:
    mov al, byte[cmd_symbol]

    cmp al, 'c'
    je cache_status_msg

    cmp al, 'e'
    je  en

    cmp al, 'd'
    je dis

    cmp al, 'a'
    je dis_all

en:
   call l2_en
   jmp cache_status

dis:
   call l2_dis
   jmp cache_status

dis_all:
   call cache_dis
   jmp cache_status

cache_status_msg:
    mov ah, 09h			;show message "Cache status"
    mov dx, cache_status_str
    int 21h

cache_status:
    mov eax, cr0
    and eax, CR0_CD             ; mask the bit CD (bit30)
    jz l2_chk                   ; go to L2 check, if caching system was enabled 
    mov ah, 09h			; show message "All caches are disabled"
    mov dx, all_cache_dis
    int 21h
    jmp short exit

l2_chk:
    cmp byte[ppro_cpu], 0
    je chk_bbl_cr_ctl3

    mov ecx, MSR_BBL_CR_CTL	; check MSR_BBL_CR_CTL register for Pentium PRO
    rdmsr
    test eax, 80000h		; check L2 cache status
    jnz show_l2_dis_msg
    jmp short show_l2_en_msg

chk_bbl_cr_ctl3:
    mov ecx, MSR_BBL_CR_CTL3	; check MSR_BBL_CR_CTL3 register for Pentium II+
    rdmsr
    test eax, 100h		; check L2 cache status
    jz show_l2_dis_msg

show_l2_en_msg:
    mov ah, 09h			;show message "L2 is enabled"
    mov dx, l2_en_str
    int 21h
    jmp short exit

show_l2_dis_msg:
    mov ah, 09h			; show message "L2 is disabled"
    mov dx, l2_dis_str
    int 21h
    jmp short exit

show_usage_msg:
    mov ah, 09h			; show message "usage"
    mov dx, usage_str
    int 21h
    jmp short exit

exit:
    mov ah, 09h
    mov dx, end_str
    int 21h
    mov ax, 4c00h
    int 21h


;========================subroutines=====================================

;=========================disable L2=====================================
l2_dis:
    pushad
    cli

    mov eax, cr0
    or eax, CR0_CD OR CR0_NW	; set the bits CD (bit30) and NW (bit29)
    wbinvd                   	; clearing L1/L2 caches
    mov cr0, eax             	; disable caching

    cmp byte[ppro_cpu], 0    	; was Pentium PRO or Pentium II+ CPU detected?
    je p2_l2dis

    mov ecx, MSR_BBL_CR_CTL  	; for Pentium PRO
    rdmsr
    or eax, 00080000h
    wrmsr                    	; disable L2 cache (set bit19)
    jmp short end_l2dis

p2_l2dis:
    mov ecx, MSR_BBL_CR_CTL3  	; for Pentium II+
    rdmsr
    and eax, NOT 0100h
    wrmsr                    	; disable L2 cache (clear bit8)

end_l2dis:
    mov eax, cr0
    and eax, NOT (CR0_CD OR CR0_NW)  ; clear the bits CD (bit30) and NW (bit29)
    mov cr0, eax             	; enable caching (really only L1)

    sti
    popad
    retn

;=========================enable L2=====================================
l2_en:
    pushad
    cli

    mov eax, cr0
    or eax, CR0_CD OR CR0_NW 	; set the bits CD (bit30) and NW (bit29)
    wbinvd                   	; clearing L1/L2 caches
    mov cr0, eax             	; disable caching

    cmp byte[ppro_cpu], 0    	; was Pentium PRO or Pentium II+ CPU detected?
    je p2_l2en

    mov ecx, MSR_BBL_CR_CTL  	; for Pentium PRO
    rdmsr
    and eax, NOT 00080000h
    wrmsr                    	; enable L2 cache (clear bit19)
    jmp short end_l2en

p2_l2en:
    mov ecx, MSR_BBL_CR_CTL3  	; for Pentium II+
    rdmsr
    or eax, 0100h
    wrmsr                    	; enable L2 cache (set bit8)

end_l2en:
    mov eax, cr0
    and eax, NOT (CR0_CD OR CR0_NW)  ; clear the bits CD (bit30) and NW (bit29)
    mov cr0, eax             ; enable caching (both, L1 and L2 caches)

    sti
    popad
    retn

;=========================disable all caches================================
cache_dis:
    push eax
    cli
    mov eax, cr0
    or eax, CR0_CD OR CR0_NW 	; set the bits CD (bit30) and NW (bit29)
    wbinvd                   	; clearing L1/L2 caches
    mov cr0, eax             	; disable caching
    sti
    pop eax
    retn

;=============================messages==================================
usage_str      db "Intel P6+ L2 cache control (c) 2017, modlabs.net v0.92, usage:", 0x0d, 0x0a
               db "l2ctl [ceda] c-check L2 status, e/d-enable/disable L2, a-disable all caches$"

non_intel_str		db "Non-Intel CPU detected!$"
cpu_unsupport_str	db "CPU is unsupported!$"
p4_cpu_str		db "Net Burst class (Pentium 4) CPU found!$"
l2_dis_str              db "Caching system is enabled, but L2 is disabled$"
l2_en_str               db "All caches are enabled$"
cache_status_str	db "Checking caches: $"
all_cache_dis		db "All caches are disabled$"
end_str			db 0xd, 0xa, '$'
cmd_symbol		db ?
ppro_cpu		db 0
